"""Provide all the classes we need for build HAT"""

from .color import ColorSensor
from .colordistance import ColorDistanceSensor
from .distance import DistanceSensor
from .exc import *  # noqa: F403
from .force import ForceSensor
from .hat import Hat
from .light import Light
from .matrix import Matrix
from .motors import Motor, MotorPair, PassiveMotor
from .serinterface import BuildHAT
from .wedo import MotionSensor, TiltSensor



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
from collections import OrderedDict

sys.path.insert(0, os.path.abspath('.'))

# Need this to correctly find the docstrings for python modules
sys.path.insert(0, os.path.abspath('..'))

# Common strings
PORT = 'hub'
VERSION = 'latest'

tags.add('port_' + PORT)
ports = OrderedDict((
    ('hub', 'the Build HAT'),
))

# The members of the html_context dict are available inside
# topindex.html

URL_PATTERN = '//en/%s/%s'
HTML_CONTEXT = {
    'port': PORT,
    'port_name': ports[PORT],
    'port_version': VERSION,
    'all_ports': [ (port_id, URL_PATTERN % (VERSION, port_id))
                   for port_id, port_name in ports.items() ],
    'all_versions': [ VERSION, URL_PATTERN % (VERSION, PORT) ],
    'downloads': [ ('PDF', URL_PATTERN % (VERSION,
                                          'python-%s.pdf' % PORT)) ]
}


# Specify a custom master document based on the port name
master_doc = PORT + "_index"

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx_selective_exclude.modindex_exclude',
    'sphinx_selective_exclude.eager_only',
    'sphinx_selective_exclude.search_auto_exclude',
    'sphinxcontrib.cmtinc-buildhat',
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['templates']

# The suffix of source filenames.
source_suffix = '.rst'

# General information about the project.
project = 'Raspberry Pi Build HAT'
copyright = '''2020-2021 - Raspberry Pi Foundation;
2017-2020 - LEGO System A/S - Aastvej 1, 7190 Billund, DK.'''

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# We don't follow "The short X.Y version" vs "The full version, including alpha/beta/rc tags"
# breakdown, so use the same version identifier for both to avoid confusion.
with open(os.path.join(os.path.dirname(__file__),'../VERSION')) as versionf:
    version = versionf.read().strip()

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['build', '.venv']
exclude_patterns += ['pending']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'any'

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# Global include files. Sphinx docs suggest using rst_epilog in preference
# of rst_prolog, so we follow. Absolute paths below mean "from the base
# of the doctree".
rst_epilog = """
.. include:: /templates/replace.inc
"""

# -- Options for HTML output ----------------------------------------------

# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    try:
        import sphinx_rtd_theme
        extensions.append("sphinx_rtd_theme")
        html_theme = 'sphinx_rtd_theme'
        html_theme_path = [sphinx_rtd_theme.get_html_theme_path(), '.']
    except:
        html_theme = 'default'
        html_theme_path = ['.']
else:
    html_theme_path = ['.']

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'build/html/_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%d %b %Y'

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {"index": "topindex.html"}

# Output file base name for HTML help builder.
htmlhelp_basename = 'BuildHATdoc'



"""HAT handling functionality"""

from .devices import Device


class Hat:
    """Allows enumeration of devices which are connected to the hat"""

    def __init__(self, device=None, debug=False):
        """Hat

        :param device: Optional string containing path to Build HAT serial device
        :param debug: Optional boolean to log debug information
        """
        self.led_status = -1
        if device is None:
            Device._setup(debug=debug)
        else:
            Device._setup(device=device, debug=debug)

    def get(self):
        """Get devices which are connected or disconnected

        :return: Dictionary of devices
        :rtype: dict
        """
        devices = {}
        for i in range(4):
            name = Device.UNKNOWN_DEVICE
            if Device._instance.connections[i].typeid in Device._device_names:
                name = Device._device_names[Device._instance.connections[i].typeid][0]
                desc = Device._device_names[Device._instance.connections[i].typeid][1]
            elif Device._instance.connections[i].typeid == -1:
                name = Device.DISCONNECTED_DEVICE
                desc = ''
            devices[chr(ord('A') + i)] = {"typeid": Device._instance.connections[i].typeid,
                                          "connected": Device._instance.connections[i].connected,
                                          "name": name,
                                          "description": desc}
        return devices

    def get_vin(self):
        """Get the voltage present on the input power jack

        :return: Voltage on the input power jack
        :rtype: float
        """
        Device._instance.write(b"vin\r")
        with Device._instance.vincond:
            Device._instance.vincond.wait()

        return Device._instance.vin

    def _set_led(self, intmode):
        if isinstance(intmode, int) and intmode >= -1 and intmode <= 3:
            self.led_status = intmode
            Device._instance.write(f"ledmode {intmode}\r".encode())

    def set_leds(self, color="voltage"):
        """Set the two LEDs on or off on the BuildHAT.

        By default the color depends on the input voltage with green being nominal at around 8V
        (The fastest time the LEDs can be perceptually toggled is around 0.025 seconds)

        :param color: orange, green, both, off, or voltage (default)
        """
        if color == "orange":
            self._set_led(1)
        elif color == "green":
            self._set_led(2)
        elif color == "both":
            self._set_led(3)
        elif color == "off":
            self._set_led(0)
        elif color == "voltage":
            self._set_led(-1)
        else:
            return

    def orange_led(self, status=True):
        """Turn the BuildHAT's orange LED on or off

        :param status: True to turn it on, False to turn it off
        """
        if status:
            if self.led_status == 3 or self.led_status == 1:
                # already on
                return
            elif self.led_status == 2:
                self._set_led(3)
            # off or default
            else:
                self._set_led(1)
        else:
            if self.led_status == 1 or self.led_status == -1:
                self._set_led(0)
            elif self.led_status == 3:
                self._set_led(2)

    def green_led(self, status=True):
        """Turn the BuildHAT's green LED on or off

        :param status: True to turn it on, False to turn it off
        """
        if status:
            if self.led_status == 3 or self.led_status == 2:
                # already on
                return
            elif self.led_status == 1:
                self._set_led(3)
            # off or default
            else:
                self._set_led(2)
        else:
            if self.led_status == 2 or self.led_status == -1:
                self._set_led(0)
            elif self.led_status == 3:
                self._set_led(1)

    def _close(self):
        Device._instance.shutdown()



"""Example for distance sensor"""

from signal import pause

from buildhat import DistanceSensor, Motor

motor = Motor('A')
dist = DistanceSensor('D', threshold_distance=100)

print("Wait for in range")
dist.wait_for_in_range(50)
motor.run_for_rotations(1)

print("Wait for out of range")
dist.wait_for_out_of_range(100)
motor.run_for_rotations(2)


def handle_in(distance):
    """Within range

    :param distance: Distance
    """
    print("in range", distance)


def handle_out(distance):
    """Out of range

    :param distance: Distance
    """
    print("out of range", distance)


dist.when_in_range = handle_in
dist.when_out_of_range = handle_out
pause()



"""Example driving LED matrix"""

import random
import time

from buildhat import Matrix

matrix = Matrix('C')

matrix.clear(("red", 10))
time.sleep(1)

matrix.clear()
time.sleep(1)

matrix.set_pixel((0, 0), ("blue", 10))
matrix.set_pixel((2, 2), ("red", 10))
time.sleep(1)

while True:
    out = [[(int(random.uniform(0, 9)), 10) for x in range(3)] for y in range(3)]
    matrix.set_pixels(out)
    time.sleep(0.1)



"""Light device handling functionality"""

from .devices import Device
from .exc import LightError


class Light(Device):
    """Light

    Use on()/off() functions to turn lights on/off

    :param port: Port of device
    :raises DeviceError: Occurs if there is no light attached to port
    """

    def __init__(self, port):
        """
        Initialise light

        :param port: Port of device
        """
        super().__init__(port)

    def brightness(self, brightness):
        """
        Brightness of LEDs

        :param brightness: Brightness argument 0 to 100
        :raises LightError: Occurs if invalid brightness passed
        """
        if not (brightness >= 0 and brightness <= 100):
            raise LightError("Need brightness arg, of 0 to 100")
        self._write(f"port {self.port} ; on ; plimit {brightness / 100.0}\r")



"""WeDo sensor handling functionality"""

from .devices import Device


class TiltSensor(Device):
    """Tilt sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no tilt sensor attached to port
    """

    def __init__(self, port):
        """
        Initialise tilt sensor

        :param port: Port of device
        """
        super().__init__(port)
        self.mode(0)

    def get_tilt(self):
        """
        Return the tilt from tilt sensor

        :return: Tilt from tilt sensor
        :rtype: tuple
        """
        return tuple(self.get())


class MotionSensor(Device):
    """Motion sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no motion sensor attached to port
    """

    def __init__(self, port):
        """
        Initialise motion sensor

        :param port: Port of device
        """
        super().__init__(port)
        self.mode(0)

    def get_distance(self):
        """
        Return the distance from motion sensor

        :return: Distance from motion sensor
        :rtype: int
        """
        return self.get()[0]



#
# This is a Sphinx documentation tool extension which allows to
# automatically exclude from full-text search index document
# which are not referenced via toctree::. It's intended to be
# used with toctrees conditional on only:: directive, with the
# idea being that if you didn't include it in the ToC, you don't
# want the docs being findable by search either (for example,
# because these docs contain information not pertinent to a
# particular product configuration).
#
# This extension depends on "eager_only" extension and won't work
# without it.
#
# Copyright (c) 2016 Paul Sokolovsky
# Licensed under the terms of BSD license, see LICENSE file.
#
import sphinx


org_StandaloneHTMLBuilder_index_page = None


def StandaloneHTMLBuilder_index_page(self, pagename, doctree, title):
    if pagename not in self.env.files_to_rebuild:
        if pagename != self.env.config.master_doc and 'orphan' not in self.env.metadata[pagename]:
            print("Excluding %s from full-text index because it's not referenced in ToC" % pagename)
            return
    return org_StandaloneHTMLBuilder_index_page(self, pagename, doctree, title)


def setup(app):
    global org_StandaloneHTMLBuilder_index_page
    org_StandaloneHTMLBuilder_index_page = sphinx.builders.html.StandaloneHTMLBuilder.index_page
    sphinx.builders.html.StandaloneHTMLBuilder.index_page = StandaloneHTMLBuilder_index_page



"""Example getting color values from color sensor"""

from buildhat import ColorSensor

color = ColorSensor('C')

print("HSV", color.get_color_hsv())
print("RGBI", color.get_color_rgbi())
print("Ambient", color.get_ambient_light())
print("Reflected", color.get_reflected_light())
print("Color", color.get_color())

print("Waiting for color black")
color.wait_until_color("black")
print("Found color black")

print("Waiting for color white")
color.wait_until_color("white")
print("Found color white")

while True:
    c = color.wait_for_new_color()
    print("Found new color", c)



"""Build HAT handling functionality"""

import logging
import queue
import tempfile
import threading
import time
from enum import Enum
from threading import Condition, Timer

import serial
from gpiozero import DigitalOutputDevice

from .exc import BuildHATError


class HatState(Enum):
    """Current state that hat is in"""

    OTHER = 0
    FIRMWARE = 1
    NEEDNEWFIRMWARE = 2
    BOOTLOADER = 3


class Connection:
    """Connection information for a port"""

    def __init__(self):
        """Initialise connection"""
        self.typeid = -1
        self.connected = False
        self.callit = None

    def update(self, typeid, connected, callit=None):
        """Update connection information for port

        :param typeid: Type ID of device on port
        :param connected: Whether device is connected or not
        :param callit: Callback function
        """
        self.typeid = typeid
        self.connected = connected
        self.callit = callit


def cmp(str1, str2):
    """Look for str2 in str1

    :param str1: String to look in
    :param str2: String to look for
    :return: Whether str2 exists
    """
    return str1[:len(str2)] == str2


class BuildHAT:
    """Interacts with Build HAT via UART interface"""

    CONNECTED = ": connected to active ID"
    CONNECTEDPASSIVE = ": connected to passive ID"
    DISCONNECTED = ": disconnected"
    DEVTIMEOUT = ": timeout during data phase: disconnecting"
    NOTCONNECTED = ": no device detected"
    PULSEDONE = ": pulse done"
    RAMPDONE = ": ramp done"
    FIRMWARE = "Firmware version: "
    BOOTLOADER = "BuildHAT bootloader version"
    DONE = "Done initialising ports"
    PROMPT = "BHBL>"
    RESET_GPIO_NUMBER = 4
    BOOT0_GPIO_NUMBER = 22

    def __init__(self, firmware, signature, version, device="/dev/serial0", debug=False):
        """Interact with Build HAT

        :param firmware: Firmware file
        :param signature: Signature file
        :param version: Firmware version
        :param device: Serial device to use
        :param debug: Optional boolean to log debug information
        :raises BuildHATError: Occurs if can't find HAT
        """
        self.cond = Condition()
        self.state = HatState.OTHER
        self.connections = []
        self.portcond = []
        self.pulsecond = []
        self.rampcond = []
        self.fin = False
        self.running = True
        self.vincond = Condition()
        self.vin = None
        if debug:
            tmp = tempfile.NamedTemporaryFile(suffix=".log", prefix="buildhat-", delete=False)
            logging.basicConfig(filename=tmp.name, format='%(asctime)s %(message)s',
                                level=logging.DEBUG)

        for _ in range(4):
            self.connections.append(Connection())
            self.portcond.append(Condition())
            self.pulsecond.append(Condition())
            self.rampcond.append(Condition())

        self.ser = serial.Serial(device, 115200, timeout=5)
        # Check if we're in the bootloader or the firmware
        self.write(b"version\r")

        emptydata = 0
        incdata = 0
        while True:
            line = self.read()
            if len(line) == 0:
                # Didn't receive any data
                emptydata += 1
                if emptydata > 3:
                    break
                else:
                    continue
            if cmp(line, BuildHAT.FIRMWARE):
                self.state = HatState.FIRMWARE
                ver = line[len(BuildHAT.FIRMWARE):].split(' ')
                if int(ver[0]) == version:
                    self.state = HatState.FIRMWARE
                    break
                else:
                    self.state = HatState.NEEDNEWFIRMWARE
                    break
            elif cmp(line, BuildHAT.BOOTLOADER):
                self.state = HatState.BOOTLOADER
                break
            else:
                # got other data we didn't understand - send version again
                incdata += 1
                if incdata > 5:
                    break
                else:
                    self.write(b"version\r")

        if self.state == HatState.NEEDNEWFIRMWARE:
            self.resethat()
            self.loadfirmware(firmware, signature)
        elif self.state == HatState.BOOTLOADER:
            self.loadfirmware(firmware, signature)
        elif self.state == HatState.OTHER:
            raise BuildHATError("HAT not found")

        self.cbqueue = queue.Queue()
        self.cb = threading.Thread(target=self.callbackloop, args=(self.cbqueue,))
        self.cb.daemon = True
        self.cb.start()

        # Drop timeout value to 1s
        self.ser.timeout = 1
        self.th = threading.Thread(target=self.loop, args=(self.cond, self.state == HatState.FIRMWARE, self.cbqueue))
        self.th.daemon = True
        self.th.start()

        if self.state == HatState.FIRMWARE:
            self.write(b"port 0 ; select ; port 1 ; select ; port 2 ; select ; port 3 ; select ; echo 0\r")
            self.write(b"list\r")
        elif self.state == HatState.NEEDNEWFIRMWARE or self.state == HatState.BOOTLOADER:
            self.write(b"reboot\r")

        # wait for initialisation to finish
        with self.cond:
            self.cond.wait()

    def resethat(self):
        """Reset the HAT"""
        reset = DigitalOutputDevice(BuildHAT.RESET_GPIO_NUMBER)
        boot0 = DigitalOutputDevice(BuildHAT.BOOT0_GPIO_NUMBER)
        boot0.off()
        reset.off()
        time.sleep(0.01)
        reset.on()
        time.sleep(0.01)
        boot0.close()
        reset.close()
        time.sleep(0.5)

    def loadfirmware(self, firmware, signature):
        """Load firmware

        :param firmware: Firmware to load
        :param signature: Signature to load
        """
        with open(firmware, "rb") as f:
            firm = f.read()
        with open(signature, "rb") as f:
            sig = f.read()
        self.write(b"clear\r")
        self.getprompt()
        self.write(f"load {len(firm)} {self.checksum(firm)}\r".encode())
        time.sleep(0.1)
        self.write(b"\x02", replace="0x02")
        self.write(firm, replace="--firmware file--")
        self.write(b"\x03\r", replace="0x03")
        self.getprompt()
        self.write(f"signature {len(sig)}\r".encode())
        time.sleep(0.1)
        self.write(b"\x02", replace="0x02")
        self.write(sig, replace="--signature file--")
        self.write(b"\x03\r", replace="0x03")
        self.getprompt()

    def getprompt(self):
        """Loop until prompt is found

        Need to decide what we will do, when no prompt
        """
        while True:
            line = self.read()
            if cmp(line, BuildHAT.PROMPT):
                break

    def checksum(self, data):
        """Calculate checksum from data

        :param data: Data to calculate the checksum from
        :return: Checksum that has been calculated
        """
        u = 1
        for i in range(0, len(data)):
            if (u & 0x80000000) != 0:
                u = (u << 1) ^ 0x1d872b41
            else:
                u = u << 1
            u = (u ^ data[i]) & 0xFFFFFFFF
        return u

    def write(self, data, log=True, replace=""):
        """Write data to the serial port of Build HAT

        :param data: Data to write to Build HAT
        :param log: Whether to log line or not
        :param replace: Whether to log an alternative string
        """
        self.ser.write(data)
        if not self.fin and log:
            if replace != "":
                logging.debug(f"> {replace}")
            else:
                logging.debug(f"> {data.decode('utf-8', 'ignore').strip()}")

    def read(self):
        """Read data from the serial port of Build HAT

        :return: Line that has been read
        """
        line = ""
        try:
            line = self.ser.readline().decode('utf-8', 'ignore').strip()
        except serial.SerialException:
            pass
        if line != "":
            logging.debug(f"< {line}")
        return line

    def shutdown(self):
        """Turn off the Build HAT devices"""
        if not self.fin:
            self.fin = True
            self.running = False
            self.th.join()
            self.cbqueue.put(())
            self.cb.join()
            turnoff = ""
            for p in range(4):
                conn = self.connections[p]
                if conn.typeid != 64:
                    turnoff += f"port {p} ; pwm ; coast ; off ;"
                else:
                    hexstr = ' '.join(f'{h:x}' for h in [0xc2, 0, 0, 0, 0, 0, 0, 0, 0, 0])
                    self.write(f"port {p} ; write1 {hexstr}\r".encode())
            self.write(f"{turnoff}\r".encode())
            self.write(b"port 0 ; select ; port 1 ; select ; port 2 ; select ; port 3 ; select ; echo 0\r")

    def callbackloop(self, q):
        """Event handling for callbacks

        :param q: Queue of callback events
        """
        while self.running:
            cb = q.get()
            # Test for empty tuple, which should only be passed when
            # we're shutting down
            if len(cb) == 0:
                continue
            if not cb[0]._alive:
                continue
            cb[0]()(cb[1])
            q.task_done()

    def loop(self, cond, uselist, q):
        """Event handling for Build HAT

        :param cond: Condition used to block user's script till we're ready
        :param uselist: Whether we're using the HATs 'list' function or not
        :param q: Queue for callback events
        """
        count = 0
        while self.running:
            line = self.read()
            if len(line) == 0:
                continue
            if line[0] == "P" and line[2] == ":":
                portid = int(line[1])
                msg = line[2:]
                if cmp(msg, BuildHAT.CONNECTED):
                    typeid = int(line[2 + len(BuildHAT.CONNECTED):], 16)
                    self.connections[portid].update(typeid, True)
                    if typeid == 64:
                        self.write(f"port {portid} ; on\r".encode())
                    if uselist:
                        count += 1
                elif cmp(msg, BuildHAT.CONNECTEDPASSIVE):
                    typeid = int(line[2 + len(BuildHAT.CONNECTEDPASSIVE):], 16)
                    self.connections[portid].update(typeid, True)
                    if uselist:
                        count += 1
                elif cmp(msg, BuildHAT.DISCONNECTED):
                    self.connections[portid].update(-1, False)
                elif cmp(msg, BuildHAT.DEVTIMEOUT):
                    self.connections[portid].update(-1, False)
                elif cmp(msg, BuildHAT.NOTCONNECTED):
                    self.connections[portid].update(-1, False)
                    if uselist:
                        count += 1
                elif cmp(msg, BuildHAT.RAMPDONE):
                    with self.rampcond[portid]:
                        self.rampcond[portid].notify()
                elif cmp(msg, BuildHAT.PULSEDONE):
                    with self.pulsecond[portid]:
                        self.pulsecond[portid].notify()

            if uselist and count == 4:
                with cond:
                    uselist = False
                    cond.notify()

            if not uselist and cmp(line, BuildHAT.DONE):
                def runit():
                    with cond:
                        cond.notify()
                t = Timer(8.0, runit)
                t.start()

            if line[0] == "P" and (line[2] == "C" or line[2] == "M"):
                portid = int(line[1])
                data = line[5:].split(" ")
                newdata = []
                for d in data:
                    if "." in d:
                        newdata.append(float(d))
                    else:
                        if d != "":
                            newdata.append(int(d))
                callit = self.connections[portid].callit
                if callit is not None:
                    q.put((callit, newdata))
                self.connections[portid].data = newdata
                with self.portcond[portid]:
                    self.portcond[portid].notify()

            if len(line) >= 5 and line[1] == "." and line.endswith(" V"):
                vin = float(line.split(" ")[0])
                self.vin = vin
                with self.vincond:
                    self.vincond.notify()



"""Example for pair of motors"""

from buildhat import MotorPair

pair = MotorPair('C', 'D')
pair.set_default_speed(20)
pair.run_for_rotations(2)

pair.run_for_rotations(1, speedl=100, speedr=20)

pair.run_to_position(20, 100, speed=20)



#
# This is a Sphinx documentation tool extension which allows to
# exclude some Python modules from the generated indexes. Modules
# are excluded both from "modindex" and "genindex" index tables
# (in the latter case, all members of a module are excluded).
# To control exclusion, set "modindex_exclude" variable in Sphinx
# conf.py to the list of modules to exclude. Note: these should be
# modules (as defined by py:module directive, not just raw filenames).
# This extension works by monkey-patching Sphinx core, so potentially
# may not work with untested Sphinx versions. It tested to work with
# 1.2.2 and 1.4.2
#
# Copyright (c) 2016 Paul Sokolovsky
# Licensed under the terms of BSD license, see LICENSE file.
#
import sphinx


#org_PythonModuleIndex_generate = None
org_PyObject_add_target_and_index = None
org_PyModule_run = None

EXCLUDES = {}

# No longer used, PyModule_run() monkey-patch does all the job
def PythonModuleIndex_generate(self, docnames=None):
    docnames = []
    excludes = self.domain.env.config['modindex_exclude']
    for modname, (docname, synopsis, platforms, deprecated) in self.domain.data['modules'].items():
        #print(docname)
        if modname not in excludes:
            docnames.append(docname)

    return org_PythonModuleIndex_generate(self, docnames)


def PyObject_add_target_and_index(self, name_cls, sig, signode):
    if hasattr(self.env, "ref_context"):
        # Sphinx 1.4
        ref_context = self.env.ref_context
    else:
        # Sphinx 1.2
        ref_context = self.env.temp_data
    modname = self.options.get(
        'module', ref_context.get('py:module'))
    #print("*", modname, name_cls)
    if modname in self.env.config['modindex_exclude']:
        return None
    return org_PyObject_add_target_and_index(self, name_cls, sig, signode)


def PyModule_run(self):
    env = self.state.document.settings.env
    modname = self.arguments[0].strip()
    excl = env.config['modindex_exclude']
    if modname in excl:
        self.options['noindex'] = True
        EXCLUDES.setdefault(modname, []).append(env.docname)
    return org_PyModule_run(self)


def setup(app):
    app.add_config_value('modindex_exclude', [], 'html')

#    global org_PythonModuleIndex_generate
#    org_PythonModuleIndex_generate = sphinx.domains.python.PythonModuleIndex.generate
#    sphinx.domains.python.PythonModuleIndex.generate = PythonModuleIndex_generate

    global org_PyObject_add_target_and_index
    org_PyObject_add_target_and_index = sphinx.domains.python.PyObject.add_target_and_index
    sphinx.domains.python.PyObject.add_target_and_index = PyObject_add_target_and_index

    global org_PyModule_run
    org_PyModule_run = sphinx.domains.python.PyModule.run
    sphinx.domains.python.PyModule.run = PyModule_run



"""Matrix device handling functionality"""

from .devices import Device
from .exc import MatrixError


class Matrix(Device):
    """LED Matrix

    :param port: Port of device
    :raises DeviceError: Occurs if there is no LED matrix attached to port
    """

    def __init__(self, port):
        """Initialise matrix

        :param port: Port of device
        """
        super().__init__(port)
        self.on()
        self.mode(2)
        self._matrix = [[(0, 0) for x in range(3)] for y in range(3)]

    def set_pixels(self, matrix, display=True):
        """Write pixel data to LED matrix

        :param matrix: 3x3 list of tuples, with colour (0–10) and brightness (0–10) (see example for more detail)
        :param display: Whether to update matrix or not
        :raises MatrixError: Occurs if invalid matrix height/width provided
        """
        if len(matrix) != 3:
            raise MatrixError("Incorrect matrix height")
        for x in range(3):
            if len(matrix[x]) != 3:
                raise MatrixError("Incorrect matrix width")
            for y in range(3):
                matrix[x][y] = Matrix.normalize_pixel(matrix[x][y])  # pylint: disable=too-many-function-args
        self._matrix = matrix
        if display:
            self._output()

    def _output(self):
        out = [0xc2]
        for x in range(3):
            for y in range(3):
                out.append((self._matrix[x][y][1] << 4) | self._matrix[x][y][0])
        self.select()
        self._write1(out)
        self.deselect()

    @staticmethod
    def strtocolor(colorstr):
        """Return the BuldHAT's integer representation of a color string

        :param colorstr: str of a valid color
        :return: (0-10) representing the color
        :rtype: int
        :raises MatrixError: Occurs if invalid color specified
        """
        if colorstr == "pink":
            return 1
        elif colorstr == "lilac":
            return 2
        elif colorstr == "blue":
            return 3
        elif colorstr == "cyan":
            return 4
        elif colorstr == "turquoise":
            return 5
        elif colorstr == "green":
            return 6
        elif colorstr == "yellow":
            return 7
        elif colorstr == "orange":
            return 8
        elif colorstr == "red":
            return 9
        elif colorstr == "white":
            return 10
        elif colorstr == "":
            return 0
        raise MatrixError("Invalid color specified")

    @staticmethod
    def normalize_pixel(pixel):
        """Validate a pixel tuple (color, brightness) and convert string colors to integers

        :param pixel: tuple of colour (0–10) or string (ie:"red") and brightness (0–10)
        :return: (color, brightness) integers
        :rtype: tuple
        :raises MatrixError: Occurs if invalid pixel specified
        """
        if isinstance(pixel, tuple):
            c, brightness = pixel  # pylint: disable=unpacking-non-sequence
            if isinstance(c, str):
                c = Matrix.strtocolor(c)  # pylint: disable=too-many-function-args
            if not (isinstance(brightness, int) and isinstance(c, int)):
                raise MatrixError("Invalid pixel specified")
            if not (brightness >= 0 and brightness <= 10):
                raise MatrixError("Invalid brightness value specified")
            if not (c >= 0 and c <= 10):
                raise MatrixError("Invalid pixel color specified")
            return (c, brightness)
        else:
            raise MatrixError("Invalid pixel specified")

    @staticmethod
    def validate_coordinate(coord):
        """Validate an x,y coordinate for the 3x3 Matrix

        :param coord: tuple of 0-2 for the X coordinate and 0-2 for the Y coordinate
        :raises MatrixError: Occurs if invalid coordinate specified
        """
        # pylint: disable=unsubscriptable-object
        if isinstance(coord, tuple):
            if not (isinstance(coord[0], int) and isinstance(coord[1], int)):
                raise MatrixError("Invalid coord specified")
            elif coord[0] > 2 or coord[0] < 0 or coord[1] > 2 or coord[1] < 0:
                raise MatrixError("Invalid coord specified")
        else:
            raise MatrixError("Invalid coord specified")

    def clear(self, pixel=None):
        """Clear matrix or set all as the same pixel

        :param pixel: tuple of colour (0–10) or string and brightness (0–10)
        """
        if pixel is None:
            self._matrix = [[(0, 0) for x in range(3)] for y in range(3)]
        else:
            color = Matrix.normalize_pixel(pixel)  # pylint: disable=too-many-function-args
            self._matrix = [[color for x in range(3)] for y in range(3)]
        self._output()

    def off(self):
        """Pretends to turn matrix off

        Never send the "off" command to the port a Matrix is connected to
        Instead, just turn all the pixels off
        """
        self.clear()

    def level(self, level):
        """Use the matrix as a "level" meter from 0-9

        (The level meter is expressed in green which seems to be unchangeable)

        :param level: The height of the bar graph, 0-9
        :raises MatrixError: Occurs if invalid level specified
        """
        if not isinstance(level, int):
            raise MatrixError("Invalid level, not integer")
        if not (level >= 0 and level <= 9):
            raise MatrixError("Invalid level specified")
        self.mode(0)
        self.select()
        self._write1([0xc0, level])
        self.mode(2)  # The rest of the Matrix code seems to expect this to be always set
        self.deselect()

    def set_transition(self, transition):
        """Set the transition mode between pixels

        Use display=False on set_pixel() or use set_pixels() to achieve desired
        results with transitions.

        Setting a new transition mode will wipe the screen and interrupt any
        running transition.

        Mode 0: No transition, immediate pixel drawing

        Mode 1: Right-to-left wipe in/out

        If the timing between writing new matrix pixels is less than one second
        the transition will clip columns of pixels from the right.

        Mode 2: Fade-in/Fade-out

        The fade in and fade out take about 2.2 seconds for full fade effect.
        Waiting less time between setting new pixels will result in a faster
        fade which will cause the fade to "pop" in brightness.

        :param transition: Transition mode (0-2)
        :raises MatrixError: Occurs if invalid transition
        """
        if not isinstance(transition, int):
            raise MatrixError("Invalid transition, not integer")
        if not (transition >= 0 and transition <= 2):
            raise MatrixError("Invalid transition specified")
        self.mode(3)
        self.select()
        self._write1([0xc3, transition])
        self.mode(2)  # The rest of the Matrix code seems to expect this to be always set
        self.deselect()

    def set_pixel(self, coord, pixel, display=True):
        """Write pixel to coordinate

        :param coord: (0,0) to (2,2)
        :param pixel: tuple of colour (0–10) or string and brightness (0–10)
        :param display: Whether to update matrix or not
        """
        color = Matrix.normalize_pixel(pixel)  # pylint: disable=too-many-function-args
        Matrix.validate_coordinate(coord)   # pylint: disable=too-many-function-args
        x, y = coord
        self._matrix[x][y] = color
        if display:
            self._output()



"""Example driving motors"""

import time

from buildhat import Motor

motor = Motor('A')
motorb = Motor('B')


def handle_motor(speed, pos, apos):
    """Motor data

    :param speed: Speed of motor
    :param pos: Position of motor
    :param apos: Absolute position of motor
    """
    print("Motor", speed, pos, apos)


motor.when_rotated = handle_motor
motor.set_default_speed(50)

print("Run for degrees 360")
motor.run_for_degrees(360)
time.sleep(3)

print("Run for degrees -360")
motor.run_for_degrees(-360)
time.sleep(3)

print("Start motor")
motor.start()
time.sleep(3)
print("Stop motor")
motor.stop()
time.sleep(1)

print("Run for degrees - 180")
motor.run_for_degrees(180)
time.sleep(3)

print("Run for degrees - 90")
motor.run_for_degrees(90)
time.sleep(3)

print("Run for rotations - 2")
motor.run_for_rotations(2)
time.sleep(3)

print("Run for seconds - 5")
motor.run_for_seconds(5)
time.sleep(3)

print("Run both")
motor.run_for_seconds(5, blocking=False)
motorb.run_for_seconds(5, blocking=False)
time.sleep(10)

print("Run to position -90")
motor.run_to_position(-90)
time.sleep(3)

print("Run to position 90")
motor.run_to_position(90)
time.sleep(3)

print("Run to position 180")
motor.run_to_position(180)
time.sleep(3)



"""Force sensor handling functionality"""

from threading import Condition

from .devices import Device


class ForceSensor(Device):
    """Force sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no force sensor attached to port
    """

    def __init__(self, port, threshold_force=1):
        """Initialise force sensor

        :param port: Port of device
        :param threshold_force: Optional
        """
        super().__init__(port)
        self.mode([(0, 0), (1, 0), (3, 0)])
        self._when_pressed = None
        self._when_released = None
        self._fired_pressed = False
        self._fired_released = False
        self._cond_force = Condition()
        self._threshold_force = threshold_force

    def _intermediate(self, data):
        with self._cond_force:
            self._data = data[0]
            self._cond_force.notify()
        if data[0] >= self.threshold_force and not self._fired_pressed:
            if self._when_pressed is not None:
                self._when_pressed(data[0])
            self._fired_pressed = True
            self._fired_released = False
        if data[0] < self.threshold_force and not self._fired_released:
            if self._when_released is not None:
                self._when_released(data[0])
            self._fired_pressed = False
            self._fired_released = True

    @property
    def threshold_force(self):
        """Threshold force

        :getter: Returns threshold force
        :setter: Sets threshold force
        :return: Threshold force
        """
        return self._threshold_force

    @threshold_force.setter
    def threshold_force(self, value):
        self._threshold_force = value

    def get_force(self):
        """Return the force in (N)

        :return: The force exerted on the button
        :rtype: int
        """
        return self.get()[0]

    def get_peak_force(self):
        """Get the maximum force registered since the sensor was reset

        (The sensor gets reset when the firmware is reloaded)

        :return: 0 - 100
        :rtype: int
        """
        return self.get()[2]

    def is_pressed(self):
        """Get whether the button is pressed

        :return: If button is pressed
        :rtype: bool
        """
        return self.get()[1] == 1

    @property
    def when_pressed(self):
        """Handle force events

        :getter: Returns function to be called when pressed
        :setter: Sets function to be called when pressed
        :return: Callback function
        """
        return self._when_pressed

    @when_pressed.setter
    def when_pressed(self, value):
        """Call back, when button is has pressed

        :param value: Callback function
        """
        self._when_pressed = value
        self.callback(self._intermediate)

    @property
    def when_released(self):
        """Handle force events

        :getter: Returns function to be called when released
        :setter: Sets function to be called when released
        :return: Callback function
        """
        return self._when_pressed

    @when_released.setter
    def when_released(self, value):
        """Call back, when button is has released

        :param value: Callback function
        """
        self._when_released = value
        self.callback(self._intermediate)

    def wait_until_pressed(self, force=1):
        """Wait until the button is pressed

        :param force: Optional
        """
        self.callback(self._intermediate)
        with self._cond_force:
            self._cond_force.wait()
            while self._data < force:
                self._cond_force.wait()

    def wait_until_released(self, force=0):
        """Wait until the button is released

        :param force: Optional
        """
        self.callback(self._intermediate)
        with self._cond_force:
            self._cond_force.wait()
            while self._data > force:
                self._cond_force.wait()



"""Example for color distance sensor"""

from buildhat import ColorDistanceSensor

color = ColorDistanceSensor('C')

print("Distance", color.get_distance())
print("RGBI", color.get_color_rgb())
print("Ambient", color.get_ambient_light())
print("Reflected", color.get_reflected_light())
print("Color", color.get_color())

print("Waiting for color black")
color.wait_until_color("black")
print("Found color black")

print("Waiting for color white")
color.wait_until_color("white")
print("Found color white")

while True:
    c = color.wait_for_new_color()
    print("Found new color", c)



"""Example using force sensor"""

from signal import pause

from buildhat import ForceSensor, Motor

motor = Motor('A')
button = ForceSensor('D', threshold_force=1)

print("Waiting for button to be pressed fully and released")

button.wait_until_pressed(100)
button.wait_until_released(0)

motor.run_for_rotations(1)

print("Wait for button to be pressed")

button.wait_until_pressed()
motor.run_for_rotations(2)


def handle_pressed(force):
    """Force sensor pressed

    :param force: Force value
    """
    print("pressed", force)


def handle_released(force):
    """Force sensor released

    :param force: Force value
    """
    print("released", force)


button.when_pressed = handle_pressed
button.when_released = handle_released
pause()



# -*- coding: utf-8 -*-
"""
    sphinxcontrib.cmtinc
    ~~~~~~~~~~~~~~~~~~~~~~~

    Extract comments from source files.

    See the README file for details.

    :author: Vilibald W. <vilibald@wvi.cz>
    :license: MIT, see LICENSE for details
"""

import codecs
import re
from docutils import nodes
from sphinx.util.nodes import nested_parse_with_titles
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive

re_comment = re.compile("^\s?/\*\*\**(.*)$")
re_cmtnext = re.compile("^[/\* | \*]?\**(.*)$")
re_cmtend = re.compile("(.*)(\*/)+$")


class ExtractError(Exception):
    pass


class Extractor(object):
    """
    Main extraction class
    """

    def __init__(self):
        """
        """
        self.content = ViewList("", 'comment')
        self.lineno = 0
        self.is_multiline = False

    def extract(self, source, prefix):
        """
        Process the source file and fill in the content.
        SOURCE is a fileobject.
        """
        for l in source:
            self.lineno = self.lineno + 1
            l = l.strip()
            m = re_comment.match(l)

            if (m):
                self.comment(m.group(1), source, prefix)

    def comment(self, cur, source, prefix):
        """
        Read the whole comment and strip the stars.

        CUR is currently read line and SOURCE is a fileobject
        with the source code.
        """
        self.content.append(cur.strip(), "comment")

        for line in source:
            self.lineno = self.lineno + 1
            # line = line.strip()

            if re_cmtend.match(line):
                if (not self.is_multiline):
                    break
                else:
                    continue

            if line.startswith("/*"):
                if (not self.is_multiline):
                    raise ExtractError(
                        "%d: Nested comments are not supported yet." %
                        self.lineno)
                else:
                    continue

            if line.startswith(".. "):
                self.content.append(line, "comment")
                continue

            if line.startswith("\code"):
                self.is_multiline = True
                continue

            if line.startswith("\endcode"):
                self.is_multiline = False
                continue

            if line.startswith("/"):
                line = "/" + line

            m = re_cmtnext.match(line)
            if m:
#               self.content.append("    " + m.group(1).strip(), "comment")
#               self.content.append("" + m.group(1), "comment")
                self.content.append(prefix + m.group(1), "comment")
                continue

            self.content.append(line.strip(), "comment")

        self.content.append('\n', "comment")


class CmtIncDirective(Directive):
    """
    Directive to insert comments form source file.
    """
    has_content = True
    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {}

    def run(self):
        """Run it """
        self.reporter = self.state.document.reporter
        self.env = self.state.document.settings.env
        if self.arguments:
            if self.content:
                return [
                    self.reporter.warning(
                        'include-comment directive cannot have content only '
                        'a filename argument',
                        line=self.lineno)
                ]
            rel_filename, filename = self.env.relfn2path(self.arguments[0])
#            prefix = '' if len(self.arguments) < 2 else re.sub('\'\"', '', self.arguments[1])
            prefix = ''
            self.env.note_dependency(rel_filename)

            extr = Extractor()
            f = None
            try:
                encoding = self.options.get('encoding',
                                            self.env.config.source_encoding)
                codecinfo = codecs.lookup(encoding)
                f = codecs.StreamReaderWriter(
                    open(filename, 'rb'), codecinfo[2], codecinfo[3], 'strict')
                extr.extract(f, prefix)
            except (IOError, OSError):
                return [
                    self.reporter.warning(
                        'Include file %r not found or reading it failed' %
                        filename,
                        line=self.lineno)
                ]
            except UnicodeError:
                return [
                    self.reporter.warning(
                        'Encoding %r used for reading included file %r seems to '
                        'be wrong, try giving an :encoding: option' %
                        (encoding, filename))
                ]
            except ExtractError as e:
                return [
                    self.reporter.warning(
                        'Parsing error in %s : %s' % (filename, str(e)),
                        line=self.lineno)
                ]
            finally:
                if f is not None:
                    f.close()
            self.content = extr.content
            self.content_offset = 0
            # Create a node, to be populated by `nested_parse`.
            node = nodes.paragraph()
            # Parse the directive contents.
            nested_parse_with_titles(self.state, self.content, node)
            return node.children
        else:
            return [
                self.reporter.warning(
                    'include-comment directive needs a filename argument',
                    line=self.lineno)
            ]


def setup(app):
    app.add_directive('include-comment', CmtIncDirective)


if __name__ == '__main__':
    import sys

    if len(sys.argv) < 2:
        print("Usage: cmtinc.py <file.c|cpp|h>")
        exit(1)

    ext = Extractor()
    try:
        with open(sys.argv[1], 'r') as f:
            ext.extract(f,"")
    except ExtractError as e:
        print('Extraction error in external source file %r : %s' %
              (sys.argv[1], str(e)))
    for line in ext.content:
        print(line)



"""Functionality for handling Build HAT devices"""

import os
import sys
import weakref

from .exc import DeviceError
from .serinterface import BuildHAT


class Device:
    """Creates a single instance of the buildhat for all devices to use"""

    _instance = None
    _started = 0
    _device_names = {1: ("PassiveMotor", "PassiveMotor"),
                     2: ("PassiveMotor", "PassiveMotor"),
                     8: ("Light", "Light"),                                   # 88005
                     34: ("TiltSensor", "WeDo 2.0 Tilt Sensor"),              # 45305
                     35: ("MotionSensor", "MotionSensor"),                    # 45304
                     37: ("ColorDistanceSensor", "Color & Distance Sensor"),  # 88007
                     61: ("ColorSensor", "Color Sensor"),                     # 45605
                     62: ("DistanceSensor", "Distance Sensor"),               # 45604
                     63: ("ForceSensor", "Force Sensor"),                     # 45606
                     64: ("Matrix", "3x3 Color Light Matrix"),                # 45608
                     38: ("Motor", "Medium Linear Motor"),                    # 88008
                     46: ("Motor", "Large Motor"),                            # 88013
                     47: ("Motor", "XL Motor"),                               # 88014
                     48: ("Motor", "Medium Angular Motor (Cyan)"),            # 45603
                     49: ("Motor", "Large Angular Motor (Cyan)"),             # 45602
                     65: ("Motor", "Small Angular Motor"),                    # 45607
                     75: ("Motor", "Medium Angular Motor (Grey)"),            # 88018
                     76: ("Motor", "Large Angular Motor (Grey)")}             # 88017

    _used = {0: False,
             1: False,
             2: False,
             3: False}

    UNKNOWN_DEVICE = "Unknown"
    DISCONNECTED_DEVICE = "Disconnected"

    def __init__(self, port):
        """Initialise device

        :param port: Port of device
        :raises DeviceError: Occurs if incorrect port specified or port already used
        """
        if not isinstance(port, str) or len(port) != 1:
            raise DeviceError("Invalid port")
        p = ord(port) - ord('A')
        if not (p >= 0 and p <= 3):
            raise DeviceError("Invalid port")
        if Device._used[p]:
            raise DeviceError("Port already used")
        self.port = p
        Device._setup()
        self._simplemode = -1
        self._combimode = -1
        self._typeid = self._conn.typeid
        if (
            self._typeid in Device._device_names
            and Device._device_names[self._typeid][0] != type(self).__name__  # noqa: W503
        ) or self._typeid == -1:
            raise DeviceError(f'There is not a {type(self).__name__} connected to port {port} (Found {self.name})')
        Device._used[p] = True

    @staticmethod
    def _setup(**kwargs):
        if Device._instance:
            return
        data = os.path.join(os.path.dirname(sys.modules["buildhat"].__file__), "data/")
        firm = os.path.join(data, "firmware.bin")
        sig = os.path.join(data, "signature.bin")
        ver = os.path.join(data, "version")
        vfile = open(ver)
        v = int(vfile.read())
        vfile.close()
        Device._instance = BuildHAT(firm, sig, v, **kwargs)
        weakref.finalize(Device._instance, Device._instance.shutdown)

    def __del__(self):
        """Handle deletion of device"""
        if hasattr(self, "port") and Device._used[self.port]:
            Device._used[self.port] = False
            self._conn.callit = None
            self.deselect()
            self.off()

    @staticmethod
    def name_for_id(typeid):
        """Translate integer type id to device name (python class)

        :param typeid: Type of device
        :return: Name of device
        """
        if typeid in Device._device_names:
            return Device._device_names[typeid][0]
        else:
            return Device.UNKNOWN_DEVICE

    @staticmethod
    def desc_for_id(typeid):
        """Translate integer type id to something more descriptive than the device name

        :param typeid: Type of device
        :return: Description of device
        """
        if typeid in Device._device_names:
            return Device._device_names[typeid][1]
        else:
            return Device.UNKNOWN_DEVICE

    @property
    def _conn(self):
        return Device._instance.connections[self.port]

    @property
    def connected(self):
        """Whether device is connected or not

        :return: Connection status
        """
        return self._conn.connected

    @property
    def typeid(self):
        """Type ID of device

        :return: Type ID
        """
        return self._typeid

    @property
    def typeidcur(self):
        """Type ID currently present

        :return: Type ID
        """
        return self._conn.typeid

    @property
    def _hat(self):
        """Hat instance

        :return: Hat instance
        """
        return Device._instance

    @property
    def name(self):
        """Determine name of device on port

        :return: Device name
        """
        if not self.connected:
            return Device.DISCONNECTED_DEVICE
        elif self.typeidcur in self._device_names:
            return self._device_names[self.typeidcur][0]
        else:
            return Device.UNKNOWN_DEVICE

    @property
    def description(self):
        """Device on port info

        :return: Device description
        """
        if not self.connected:
            return Device.DISCONNECTED_DEVICE
        elif self.typeidcur in self._device_names:
            return self._device_names[self.typeidcur][1]
        else:
            return Device.UNKNOWN_DEVICE

    def isconnected(self):
        """Whether it is connected or not

        :raises DeviceError: Occurs if device no longer the same
        """
        if not self.connected:
            raise DeviceError("No device found")
        if self.typeid != self.typeidcur:
            raise DeviceError("Device has changed")

    def reverse(self):
        """Reverse polarity"""
        self._write(f"port {self.port} ; plimit 1 ; set -1\r")

    def get(self):
        """Extract information from device

        :return: Data from device
        :raises DeviceError: Occurs if device not in valid mode
        """
        self.isconnected()
        idx = -1
        if self._simplemode != -1:
            idx = self._simplemode
        elif self._combimode != -1:
            idx = self._combimode
        else:
            raise DeviceError("Not in simple or combimode")
        self._write(f"port {self.port} ; selonce {idx}\r")
        # wait for data
        with Device._instance.portcond[self.port]:
            Device._instance.portcond[self.port].wait()
        return self._conn.data

    def mode(self, modev):
        """Set combimode or simple mode

        :param modev: List of tuples for a combimode, or integer for simple mode
        """
        self.isconnected()
        if isinstance(modev, list):
            self._combimode = 0
            modestr = ""
            for t in modev:
                modestr += f"{t[0]} {t[1]} "
            self._write(f"port {self.port} ; combi {self._combimode} {modestr}\r")
            self._simplemode = -1
        else:
            # Remove combi mode
            if self._combimode != -1:
                self._write(f"port {self.port} ; combi {self._combimode}\r")
            self._combimode = -1
            self._simplemode = int(modev)

    def select(self):
        """Request data from mode

        :raises DeviceError: Occurs if device not in valid mode
        """
        self.isconnected()
        if self._simplemode != -1:
            idx = self._simplemode
        elif self._combimode != -1:
            idx = self._combimode
        else:
            raise DeviceError("Not in simple or combimode")
        self._write(f"port {self.port} ; select {idx}\r")

    def on(self):
        """Turn on sensor"""
        self._write(f"port {self.port} ; plimit 1 ; on\r")

    def off(self):
        """Turn off sensor"""
        self._write(f"port {self.port} ; off\r")

    def deselect(self):
        """Unselect data from mode"""
        self._write(f"port {self.port} ; select\r")

    def _write(self, cmd):
        self.isconnected()
        Device._instance.write(cmd.encode())

    def _write1(self, data):
        hexstr = ' '.join(f'{h:x}' for h in data)
        self._write(f"port {self.port} ; write1 {hexstr}\r")

    def callback(self, func):
        """Set callback function

        :param func: Callback function
        """
        if func is not None:
            self.select()
        else:
            self.deselect()
        if func is None:
            self._conn.callit = None
        else:
            self._conn.callit = weakref.WeakMethod(func)



"""Distance sensor handling functionality"""

from threading import Condition

from .devices import Device
from .exc import DistanceSensorError


class DistanceSensor(Device):
    """Distance sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no distance sensor attached to port
    """

    def __init__(self, port, threshold_distance=100):
        """
        Initialise distance sensor

        :param port: Port of device
        :param threshold_distance: Optional
        """
        super().__init__(port)
        self.on()
        self.mode(0)
        self._cond_data = Condition()
        self._when_in_range = None
        self._when_out_of_range = None
        self._fired_in = False
        self._fired_out = False
        self._threshold_distance = threshold_distance
        self._distance = -1

    def _intermediate(self, data):
        self._distance = data[0]
        if self._distance != -1 and self._distance < self.threshold_distance and not self._fired_in:
            if self._when_in_range is not None:
                self._when_in_range(data[0])
            self._fired_in = True
            self._fired_out = False
        if self._distance != -1 and self._distance > self.threshold_distance and not self._fired_out:
            if self._when_out_of_range is not None:
                self._when_out_of_range(data[0])
            self._fired_in = False
            self._fired_out = True
        with self._cond_data:
            self._data = data[0]
            self._cond_data.notify()

    @property
    def distance(self):
        """
        Obtain previously stored distance

        :getter: Returns distance
        :return: Stored distance
        """
        return self._distance

    @property
    def threshold_distance(self):
        """
        Threshold distance value

        :getter: Returns threshold distance
        :setter: Sets threshold distance
        :return: Threshold distance
        """
        return self._threshold_distance

    @threshold_distance.setter
    def threshold_distance(self, value):
        self._threshold_distance = value

    def get_distance(self):
        """
        Return the distance from ultrasonic sensor to object

        :return: Distance from ultrasonic sensor
        :rtype: int
        """
        dist = self.get()[0]
        return dist

    @property
    def when_in_range(self):
        """
        Handle motion events

        :getter: Returns function to be called when in range
        :setter: Sets function to be called when in range
        :return: In range callback
        """
        return self._when_in_range

    @when_in_range.setter
    def when_in_range(self, value):
        """Call back, when distance in range

        :param value: In range callback
        """
        self._when_in_range = value
        self.callback(self._intermediate)

    @property
    def when_out_of_range(self):
        """
        Handle motion events

        :getter: Returns function to be called when out of range
        :setter: Sets function to be called when out of range
        :return: Out of range callback
        """
        return self._when_out_of_range

    @when_out_of_range.setter
    def when_out_of_range(self, value):
        """Call back, when distance out of range

        :param value: Out of range callback
        """
        self._when_out_of_range = value
        self.callback(self._intermediate)

    def wait_for_out_of_range(self, distance):
        """Wait until object is farther than specified distance

        :param distance: Distance
        """
        self.callback(self._intermediate)
        with self._cond_data:
            self._cond_data.wait()
            while self._data < distance:
                self._cond_data.wait()

    def wait_for_in_range(self, distance):
        """Wait until object is closer than specified distance

        :param distance: Distance
        """
        self.callback(self._intermediate)
        with self._cond_data:
            self._cond_data.wait()
            while self._data == -1 or self._data > distance:
                self._cond_data.wait()

    def eyes(self, *args):
        """
        Brightness of LEDs on sensor

        (Sensor Right Upper, Sensor Left Upper, Sensor Right Lower, Sensor Left Lower)

        :param args: Four brightness arguments of 0 to 100
        :raises DistanceSensorError: Occurs if invalid brightness passed
        """
        out = [0xc5]
        if len(args) != 4:
            raise DistanceSensorError("Need 4 brightness args, of 0 to 100")
        for v in args:
            if not (v >= 0 and v <= 100):
                raise DistanceSensorError("Need 4 brightness args, of 0 to 100")
            out += [v]
        self._write1(out)

    def on(self):
        """Turn on the sensor"""
        self._write(f"port {self.port} ; set -1\r")



"""Exceptions for all build HAT classes"""


class DistanceSensorError(Exception):
    """Error raised when invalid arguments passed to distance sensor functions"""


class MatrixError(Exception):
    """Error raised when invalid arguments passed to matrix functions"""


class LightError(Exception):
    """Error raised when invalid arguments passed to light functions"""


class MotorError(Exception):
    """Error raised when invalid arguments passed to motor functions"""


class BuildHATError(Exception):
    """Error raised when HAT not found"""


class DeviceError(Exception):
    """Error raised when there is a Device issue"""



"""Motor device handling functionality"""

import threading
import time
from collections import deque
from enum import Enum
from threading import Condition

from .devices import Device
from .exc import MotorError


class PassiveMotor(Device):
    """Passive Motor device

    :param port: Port of device
    :raises DeviceError: Occurs if there is no passive motor attached to port
    """

    def __init__(self, port):
        """Initialise motor

        :param port: Port of device
        """
        super().__init__(port)
        self._default_speed = 20
        self._currentspeed = 0
        self.plimit(0.7)
        self.bias(0.3)

    def set_default_speed(self, default_speed):
        """Set the default speed of the motor

        :param default_speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs if invalid speed passed
        """
        if not (default_speed >= -100 and default_speed <= 100):
            raise MotorError("Invalid Speed")
        self._default_speed = default_speed

    def start(self, speed=None):
        """Start motor

        :param speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs if invalid speed passed
        """
        if self._currentspeed == speed:
            # Already running at this speed, do nothing
            return

        if speed is None:
            speed = self._default_speed
        else:
            if not (speed >= -100 and speed <= 100):
                raise MotorError("Invalid Speed")
        self._currentspeed = speed
        cmd = f"port {self.port} ; pwm ; set {speed / 100}\r"
        self._write(cmd)

    def stop(self):
        """Stop motor"""
        cmd = f"port {self.port} ; off\r"
        self._write(cmd)
        self._currentspeed = 0

    def plimit(self, plimit):
        """Limit power

        :param plimit: Value 0 to 1
        :raises MotorError: Occurs if invalid plimit value passed
        """
        if not (plimit >= 0 and plimit <= 1):
            raise MotorError("plimit should be 0 to 1")
        self._write(f"port {self.port} ; plimit {plimit}\r")

    def bias(self, bias):
        """Bias motor

        :param bias: Value 0 to 1
        :raises MotorError: Occurs if invalid bias value passed
        """
        if not (bias >= 0 and bias <= 1):
            raise MotorError("bias should be 0 to 1")
        self._write(f"port {self.port} ; bias {bias}\r")


class MotorRunmode(Enum):
    """Current mode motor is in"""

    NONE = 0
    FREE = 1
    DEGREES = 2
    SECONDS = 3


class Motor(Device):
    """Motor device

    :param port: Port of device
    :raises DeviceError: Occurs if there is no motor attached to port
    """

    def __init__(self, port):
        """Initialise motor

        :param port: Port of device
        """
        super().__init__(port)
        self.default_speed = 20
        self._currentspeed = 0
        if self._typeid in {38}:
            self.mode([(1, 0), (2, 0)])
            self._combi = "1 0 2 0"
            self._noapos = True
        else:
            self.mode([(1, 0), (2, 0), (3, 0)])
            self._combi = "1 0 2 0 3 0"
            self._noapos = False
        self.plimit(0.7)
        self.bias(0.3)
        self._release = True
        self._bqueue = deque(maxlen=5)
        self._cvqueue = Condition()
        self.when_rotated = None
        self._oldpos = None
        self._runmode = MotorRunmode.NONE

    def set_default_speed(self, default_speed):
        """Set the default speed of the motor

        :param default_speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs if invalid speed passed
        """
        if not (default_speed >= -100 and default_speed <= 100):
            raise MotorError("Invalid Speed")
        self.default_speed = default_speed

    def run_for_rotations(self, rotations, speed=None, blocking=True):
        """Run motor for N rotations

        :param rotations: Number of rotations
        :param speed: Speed ranging from -100 to 100
        :param blocking: Whether call should block till finished
        :raises MotorError: Occurs if invalid speed passed
        """
        self._runmode = MotorRunmode.DEGREES
        if speed is None:
            self.run_for_degrees(int(rotations * 360), self.default_speed, blocking)
        else:
            if not (speed >= -100 and speed <= 100):
                raise MotorError("Invalid Speed")
            self.run_for_degrees(int(rotations * 360), speed, blocking)

    def _run_for_degrees(self, degrees, speed):
        self._runmode = MotorRunmode.DEGREES
        mul = 1
        if speed < 0:
            speed = abs(speed)
            mul = -1
        pos = self.get_position()
        newpos = ((degrees * mul) + pos) / 360.0
        pos /= 360.0
        self._run_positional_ramp(pos, newpos, speed)
        self._runmode = MotorRunmode.NONE

    def _run_to_position(self, degrees, speed, direction):
        self._runmode = MotorRunmode.DEGREES
        data = self.get()
        pos = data[1]
        if self._noapos:
            apos = pos
        else:
            apos = data[2]
        diff = (degrees - apos + 180) % 360 - 180
        newpos = (pos + diff) / 360
        v1 = (degrees - apos) % 360
        v2 = (apos - degrees) % 360
        mul = 1
        if diff > 0:
            mul = -1
        diff = sorted([diff, mul * (v2 if abs(diff) == v1 else v1)])
        if direction == "shortest":
            pass
        elif direction == "clockwise":
            newpos = (pos + diff[1]) / 360
        elif direction == "anticlockwise":
            newpos = (pos + diff[0]) / 360
        else:
            raise MotorError("Invalid direction, should be: shortest, clockwise or anticlockwise")
        # Convert current motor position to decimal rotations from preset position to match newpos units
        pos /= 360.0
        self._run_positional_ramp(pos, newpos, speed)
        self._runmode = MotorRunmode.NONE

    def _run_positional_ramp(self, pos, newpos, speed):
        """Ramp motor

        :param pos: Current motor position in decimal rotations (from preset position)
        :param newpos: New motor postion in decimal rotations (from preset position)
        :param speed: -100 to 100
        """
        # Collapse speed range to -5 to 5
        speed *= 0.05
        dur = abs((newpos - pos) / speed)
        cmd = (f"port {self.port}; combi 0 {self._combi} ; select 0 ; pid {self.port} 0 1 s4 0.0027777778 0 5 0 .1 3 ; "
               f"set ramp {pos} {newpos} {dur} 0\r")
        self._write(cmd)
        with self._hat.rampcond[self.port]:
            self._hat.rampcond[self.port].wait()
        if self._release:
            time.sleep(0.2)
            self.coast()

    def run_for_degrees(self, degrees, speed=None, blocking=True):
        """Run motor for N degrees

        Speed of 1 means 1 revolution / second

        :param degrees: Number of degrees to rotate
        :param speed: Speed ranging from -100 to 100
        :param blocking: Whether call should block till finished
        :raises MotorError: Occurs if invalid speed passed
        """
        self._runmode = MotorRunmode.DEGREES
        if speed is None:
            speed = self.default_speed
        if not (speed >= -100 and speed <= 100):
            raise MotorError("Invalid Speed")
        if not blocking:
            th = threading.Thread(target=self._run_for_degrees, args=(degrees, speed))
            th.daemon = True
            th.start()
        else:
            self._run_for_degrees(degrees, speed)

    def run_to_position(self, degrees, speed=None, blocking=True, direction="shortest"):
        """Run motor to position (in degrees)

        :param degrees: Position in degrees from -180 to 180
        :param speed: Speed ranging from 0 to 100
        :param blocking: Whether call should block till finished
        :param direction: shortest (default)/clockwise/anticlockwise
        :raises MotorError: Occurs if invalid speed or angle passed
        """
        self._runmode = MotorRunmode.DEGREES
        if speed is None:
            speed = self.default_speed
        if not (speed >= 0 and speed <= 100):
            raise MotorError("Invalid Speed")
        if degrees < -180 or degrees > 180:
            raise MotorError("Invalid angle")
        if not blocking:
            th = threading.Thread(target=self._run_to_position, args=(degrees, speed, direction))
            th.daemon = True
            th.start()
        else:
            self._run_to_position(degrees, speed, direction)

    def _run_for_seconds(self, seconds, speed):
        self._runmode = MotorRunmode.SECONDS
        cmd = (f"port {self.port} ; combi 0 {self._combi} ; select 0 ; pid {self.port} 0 0 s1 1 0 0.003 0.01 0 100; "
               f"set pulse {speed} 0.0 {seconds} 0\r")
        self._write(cmd)
        with self._hat.pulsecond[self.port]:
            self._hat.pulsecond[self.port].wait()
        if self._release:
            self.coast()
        self._runmode = MotorRunmode.NONE

    def run_for_seconds(self, seconds, speed=None, blocking=True):
        """Run motor for N seconds

        :param seconds: Time in seconds
        :param speed: Speed ranging from -100 to 100
        :param blocking: Whether call should block till finished
        :raises MotorError: Occurs when invalid speed specified
        """
        self._runmode = MotorRunmode.SECONDS
        if speed is None:
            speed = self.default_speed
        if not (speed >= -100 and speed <= 100):
            raise MotorError("Invalid Speed")
        if not blocking:
            th = threading.Thread(target=self._run_for_seconds, args=(seconds, speed))
            th.daemon = True
            th.start()
        else:
            self._run_for_seconds(seconds, speed)

    def start(self, speed=None):
        """Start motor

        :param speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs when invalid speed specified
        """
        if self._runmode == MotorRunmode.FREE:
            if self._currentspeed == speed:
                # Already running at this speed, do nothing
                return
        elif self._runmode != MotorRunmode.NONE:
            # Motor is running some other mode, wait for it to stop or stop() it yourself
            return

        if speed is None:
            speed = self.default_speed
        else:
            if not (speed >= -100 and speed <= 100):
                raise MotorError("Invalid Speed")
        cmd = f"port {self.port} ; set {speed}\r"
        if self._runmode == MotorRunmode.NONE:
            cmd = (f"port {self.port} ; combi 0 {self._combi} ; select 0 ; pid {self.port} 0 0 s1 1 0 0.003 0.01 0 100; "
                   f"set {speed}\r")
        self._runmode = MotorRunmode.FREE
        self._currentspeed = speed
        self._write(cmd)

    def stop(self):
        """Stop motor"""
        self._runmode = MotorRunmode.NONE
        self._currentspeed = 0
        self.coast()

    def get_position(self):
        """Get position of motor with relation to preset position (can be negative or positive)

        :return: Position of motor in degrees from preset position
        :rtype: int
        """
        return self.get()[1]

    def get_aposition(self):
        """Get absolute position of motor

        :return: Absolute position of motor from -180 to 180
        :rtype: int
        """
        if self._noapos:
            raise MotorError("No absolute position with this motor")
        else:
            return self.get()[2]

    def get_speed(self):
        """Get speed of motor

        :return: Speed of motor
        :rtype: int
        """
        return self.get()[0]

    @property
    def when_rotated(self):
        """
        Handle rotation events

        :getter: Returns function to be called when rotated
        :setter: Sets function to be called when rotated
        :return: Callback function
        """
        return self._when_rotated

    def _intermediate(self, data):
        if self._noapos:
            speed, pos = data
            apos = None
        else:
            speed, pos, apos = data
        if self._oldpos is None:
            self._oldpos = pos
            return
        if abs(pos - self._oldpos) >= 1:
            if self._when_rotated is not None:
                self._when_rotated(speed, pos, apos)
            self._oldpos = pos

    @when_rotated.setter
    def when_rotated(self, value):
        """Call back, when motor has been rotated

        :param value: Callback function
        """
        self._when_rotated = value
        self.callback(self._intermediate)

    def plimit(self, plimit):
        """Limit power

        :param plimit: Value 0 to 1
        :raises MotorError: Occurs if invalid plimit value passed
        """
        if not (plimit >= 0 and plimit <= 1):
            raise MotorError("plimit should be 0 to 1")
        self._write(f"port {self.port} ; plimit {plimit}\r")

    def bias(self, bias):
        """Bias motor

        :param bias: Value 0 to 1
        :raises MotorError: Occurs if invalid bias value passed
        """
        if not (bias >= 0 and bias <= 1):
            raise MotorError("bias should be 0 to 1")
        self._write(f"port {self.port} ; bias {bias}\r")

    def pwm(self, pwmv):
        """PWM motor

        :param pwmv: Value -1 to 1
        :raises MotorError: Occurs if invalid pwm value passed
        """
        if not (pwmv >= -1 and pwmv <= 1):
            raise MotorError("pwm should be -1 to 1")
        self._write(f"port {self.port} ; pwm ; set {pwmv}\r")

    def coast(self):
        """Coast motor"""
        self._write(f"port {self.port} ; coast\r")

    def float(self):
        """Float motor"""
        self.pwm(0)

    @property
    def release(self):
        """Determine if motor is released after running, so can be turned by hand

        :getter: Returns whether motor is released, so can be turned by hand
        :setter: Sets whether motor is released, so can be turned by hand
        :return: Whether motor is released, so can be turned by hand
        :rtype: bool
        """
        return self._release

    @release.setter
    def release(self, value):
        """Determine if the motor is released after running, so can be turned by hand

        :param value: Whether motor should be released, so can be turned by hand
        :type value: bool
        """
        if not isinstance(value, bool):
            raise MotorError("Must pass boolean")
        self._release = value


class MotorPair:
    """Pair of motors

    :param motora: One of the motors to drive
    :param motorb: Other motor in pair to drive
    :raises DeviceError: Occurs if there is no motor attached to port
    """

    def __init__(self, leftport, rightport):
        """Initialise pair of motors

        :param leftport: Left motor port
        :param rightport:  Right motor port
        """
        super().__init__()
        self._leftmotor = Motor(leftport)
        self._rightmotor = Motor(rightport)
        self.default_speed = 20
        self._release = True

    def set_default_speed(self, default_speed):
        """Set the default speed of the motor

        :param default_speed: Speed ranging from -100 to 100
        """
        self.default_speed = default_speed

    def run_for_rotations(self, rotations, speedl=None, speedr=None):
        """Run pair of motors for N rotations

        :param rotations: Number of rotations
        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        self.run_for_degrees(int(rotations * 360), speedl, speedr)

    def run_for_degrees(self, degrees, speedl=None, speedr=None):
        """Run pair of motors for degrees

        :param degrees: Number of degrees
        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        th1 = threading.Thread(target=self._leftmotor._run_for_degrees, args=(degrees, speedl))
        th2 = threading.Thread(target=self._rightmotor._run_for_degrees, args=(degrees, speedr))
        th1.daemon = True
        th2.daemon = True
        th1.start()
        th2.start()
        th1.join()
        th2.join()

    def run_for_seconds(self, seconds, speedl=None, speedr=None):
        """Run pair for N seconds

        :param seconds: Time in seconds
        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        th1 = threading.Thread(target=self._leftmotor._run_for_seconds, args=(seconds, speedl))
        th2 = threading.Thread(target=self._rightmotor._run_for_seconds, args=(seconds, speedr))
        th1.daemon = True
        th2.daemon = True
        th1.start()
        th2.start()
        th1.join()
        th2.join()

    def start(self, speedl=None, speedr=None):
        """Start motors

        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        self._leftmotor.start(speedl)
        self._rightmotor.start(speedr)

    def stop(self):
        """Stop motors"""
        self._leftmotor.stop()
        self._rightmotor.stop()

    def run_to_position(self, degreesl, degreesr, speed=None, direction="shortest"):
        """Run pair to position (in degrees)

        :param degreesl: Position in degrees for left motor
        :param degreesr: Position in degrees for right motor
        :param speed: Speed ranging from -100 to 100
        :param direction: shortest (default)/clockwise/anticlockwise
        """
        if speed is None:
            th1 = threading.Thread(target=self._leftmotor._run_to_position, args=(degreesl, self.default_speed, direction))
            th2 = threading.Thread(target=self._rightmotor._run_to_position, args=(degreesr, self.default_speed, direction))
        else:
            th1 = threading.Thread(target=self._leftmotor._run_to_position, args=(degreesl, speed, direction))
            th2 = threading.Thread(target=self._rightmotor._run_to_position, args=(degreesr, speed, direction))
        th1.daemon = True
        th2.daemon = True
        th1.start()
        th2.start()
        th1.join()
        th2.join()

    @property
    def release(self):
        """Determine if motors are released after running, so can be turned by hand

        :getter: Returns whether motors are released, so can be turned by hand
        :setter: Sets whether motors are released, so can be turned by hand
        :return: Whether motors are released, so can be turned by hand
        :rtype: bool
        """
        return self._release

    @release.setter
    def release(self, value):
        """Determine if motors are released after running, so can be turned by hand

        :param value: Whether motors should be released, so can be turned by hand
        :type value: bool
        """
        if not isinstance(value, bool):
            raise MotorError("Must pass boolean")
        self._release = value
        self._leftmotor.release = value
        self._rightmotor.release = value



#! /usr/bin/env python3

"""Setup file"""

# Copyright (c) 2020-2021 Raspberry Pi Foundation
#
# SPDX-License-Identifier: MIT

from setuptools import setup

with open("README.md") as readme:
    long_description = readme.read()

with open('VERSION') as versionf:
    version = versionf.read().strip()

setup(name='buildhat',
      version=version,
      description='Build HAT Python library',
      license='MIT',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Raspberry Pi Foundation',
      author_email='web@raspberrypi.org',
      url='https://github.com/RaspberryPiFoundation/python-build-hat',
      project_urls={'Bug Tracker': "https://github.com/RaspberryPiFoundation/python-build-hat/issues"},
      packages=['buildhat'],
      package_data={
          "": ["data/firmware.bin", "data/signature.bin", "data/version"],
      },
      python_requires='>=3.7',
      install_requires=['gpiozero', 'pyserial'])



#
# This is a Sphinx documentation tool extension which makes .only::
# directives be eagerly processed early in the parsing stage. This
# makes sure that content in .only:: blocks gets actually excluded
# as a typical user expects, instead of bits of information in
# these blocks leaking to documentation in various ways (e.g.,
# indexes containing entries for functions which are actually in
# .only:: blocks and thus excluded from documentation, etc.)
# Note that with this extension, you may need to completely
# rebuild a doctree when switching builders (i.e. completely
# remove _build/doctree dir between generation of HTML vs PDF
# documentation).
#
# This extension works by monkey-patching Sphinx core, so potentially
# may not work with untested Sphinx versions. It tested to work with
# 1.2.2 and 1.4.2
#
# Copyright (c) 2016 Paul Sokolovsky
# Based on idea by Andrea Cassioli:
# https://github.com/sphinx-doc/sphinx/issues/2150#issuecomment-171912290
# Licensed under the terms of BSD license, see LICENSE file.
#
import sphinx
from docutils.parsers.rst import directives


class EagerOnly(sphinx.directives.other.Only):

    def run(self, *args):
        # Evaluate the condition eagerly, and if false return no nodes right away
        env = self.state.document.settings.env
        env.app.builder.tags.add('TRUE')
        #print(repr(self.arguments[0]))
        if not env.app.builder.tags.eval_condition(self.arguments[0]):
            return []

        # Otherwise, do the usual processing
        nodes = super(EagerOnly, self).run()
        if len(nodes) == 1:
            nodes[0]['expr'] = 'TRUE'
        return nodes


def setup(app):
    directives.register_directive('only', EagerOnly)



"""Color distance sensor handling functionality"""

import math
from collections import deque
from threading import Condition

from .devices import Device


class ColorDistanceSensor(Device):
    """Color Distance sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no colordistance sensor attached to port
    """

    def __init__(self, port):
        """
        Initialise color distance sensor

        :param port: Port of device
        """
        super().__init__(port)
        self.on()
        self.mode(6)
        self.avg_reads = 4
        self._old_color = None

    def segment_color(self, r, g, b):
        """Return the color name from HSV

        :param r: Red
        :param g: Green
        :param b: Blue
        :return: Name of the color as a string
        :rtype: str
        """
        table = [("black", (0, 0, 0)),
                 ("violet", (127, 0, 255)),
                 ("blue", (0, 0, 255)),
                 ("cyan", (0, 183, 235)),
                 ("green", (0, 128, 0)),
                 ("yellow", (255, 255, 0)),
                 ("red", (255, 0, 0)),
                 ("white", (255, 255, 255))]
        near = ""
        euc = math.inf
        for itm in table:
            cur = math.sqrt((r - itm[1][0]) ** 2 + (g - itm[1][1]) ** 2 + (b - itm[1][2]) ** 2)
            if cur < euc:
                near = itm[0]
                euc = cur
        return near

    def rgb_to_hsv(self, r, g, b):
        """Convert RGB to HSV

        Based on https://www.rapidtables.com/convert/color/rgb-to-hsv.html algorithm

        :param r: Red
        :param g: Green
        :param b: Blue
        :return: HSV representation of color
        :rtype: tuple
        """
        r, g, b = r / 255.0, g / 255.0, b / 255.0
        cmax = max(r, g, b)
        cmin = min(r, g, b)
        delt = cmax - cmin
        if cmax == cmin:
            h = 0
        elif cmax == r:
            h = 60 * (((g - b) / delt) % 6)
        elif cmax == g:
            h = 60 * ((((b - r) / delt)) + 2)
        elif cmax == b:
            h = 60 * ((((r - g) / delt)) + 4)
        if cmax == 0:
            s = 0
        else:
            s = delt / cmax
        v = cmax
        return int(h), int(s * 100), int(v * 100)

    def get_color(self):
        """Return the color

        :return: Name of the color as a string
        :rtype: str
        """
        r, g, b = self.get_color_rgb()
        return self.segment_color(r, g, b)

    def get_ambient_light(self):
        """Return the ambient light

        :return: Ambient light
        :rtype: int
        """
        self.mode(4)
        readings = []
        for _ in range(self.avg_reads):
            readings.append(self.get()[0])
        return int(sum(readings) / len(readings))

    def get_reflected_light(self):
        """Return the reflected light

        :return: Reflected light
        :rtype: int
        """
        self.mode(3)
        readings = []
        for _ in range(self.avg_reads):
            readings.append(self.get()[0])
        return int(sum(readings) / len(readings))

    def get_distance(self):
        """Return the distance

        :return: Distance
        :rtype: int
        """
        self.mode(1)
        distance = self.get()[0]
        return distance

    def _clamp(self, val, small, large):
        return max(small, min(val, large))

    def _avgrgb(self, reads):
        readings = []
        for read in reads:
            read = [int((self._clamp(read[0], 0, 400) / 400) * 255),
                    int((self._clamp(read[1], 0, 400) / 400) * 255),
                    int((self._clamp(read[2], 0, 400) / 400) * 255)]
            readings.append(read)
        rgb = []
        for i in range(3):
            rgb.append(int(sum([rgb[i] for rgb in readings]) / len(readings)))
        return rgb

    def get_color_rgb(self):
        """Return the color

        :return: RGBI representation
        :rtype: list
        """
        self.mode(6)
        reads = []
        for _ in range(self.avg_reads):
            reads.append(self.get())
        return self._avgrgb(reads)

    def _cb_handle(self, lst):
        self._data.append(lst)
        if len(self._data) == self.avg_reads:
            r, g, b = self._avgrgb(self._data)
            seg = self.segment_color(r, g, b)
            if self._cmp(seg, self._color):
                with self._cond:
                    self._old_color = seg
                    self._cond.notify()

    def wait_until_color(self, color):
        """Wait until specific color

        :param color: Color to look for
        """
        self.mode(6)
        self._cond = Condition()
        self._data = deque(maxlen=self.avg_reads)
        self._color = color
        self._cmp = lambda x, y: x == y
        self.callback(self._cb_handle)
        with self._cond:
            self._cond.wait()
        self.callback(None)

    def wait_for_new_color(self):
        """Wait for new color or returns immediately if first call

        :return: Name of the color as a string
        :rtype: str
        """
        self.mode(6)
        if self._old_color is None:
            self._old_color = self.get_color()
            return self._old_color
        self._cond = Condition()
        self._data = deque(maxlen=self.avg_reads)
        self._color = self._old_color
        self._cmp = lambda x, y: x != y
        self.callback(self._cb_handle)
        with self._cond:
            self._cond.wait()
        self.callback(None)
        return self._old_color

    def on(self):
        """Turn on the sensor and LED"""
        self._write(f"port {self.port} ; plimit 1 ; set -1\r")
